/********************************************************
 * fdup -- Duplicate floppies.				*
 *							*
 * fdup -?	for usage				*
 ********************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/fd.h>
#include <sys/param.h>
#include <regex.h>

const int SECTOR_SIZE = 512;	// Size of a sector in bytes

static bool debug = false;	// If set, output debug messages
static bool format = false;	// If set, format the destination disks
static bool verify = false;	// If set, verify the disk
static bool beeps = false;	// If set, beep on changes
static bool continue_on_error = false;	// If set, ignore errors
static int dst_count = 1; 	// Number of disks to write
static int src_count = 1000;	// Number of disks to read
static bool graphic = false;	// Display things graphically
static int device_number = 0;	// Floppy number


// A list of devices to try and open to see if we can read a floppy
static char *probe_list[] = {
    "/dev/fd%d",
    "/dev/fd%dH1440",
    "/dev/fd%dh1200",
    "/dev/fd%dD360",
    "/dev/fd%dD720",
    "/dev/fd%dH360",
    "/dev/fd%dH720",
    "/dev/fd%dd360",
    "/dev/fd%dh360",
    "/dev/fd%dh720",
    "/dev/fd%d",
    NULL
};

// Name of the device to probe
static char probe_device_name[MAXPATHLEN] = "/dev/fd0";
// Name of the device to read / write 
static char actual_device_name[MAXPATHLEN] = "/dev/fd0";

static struct floppy_struct floppy_info;	/* Information on the floppy */
static long int track_size = 0;			/* Size of a track */

// The floppy structure formation (from the kernel file "floppy.c")
struct floppy_struct info_list[] = {
	{    0, 0,0, 0,0,0x00,0x00,0x00,0x00,NULL    },	/*  0 no testing    */
	{  720, 9,2,40,0,0x2A,0x02,0xDF,0x50,"d360"  }, /*  1 360KB PC      */
	{ 2400,15,2,80,0,0x1B,0x00,0xDF,0x54,"h1200" },	/*  2 1.2MB AT      */
	{  720, 9,1,80,0,0x2A,0x02,0xDF,0x50,"D360"  },	/*  3 360KB SS 3.5" */
	{ 1440, 9,2,80,0,0x2A,0x02,0xDF,0x50,"D720"  },	/*  4 720KB 3.5"    */
	{  720, 9,2,40,1,0x23,0x01,0xDF,0x50,"h360"  },	/*  5 360KB AT      */
	{ 1440, 9,2,80,0,0x23,0x01,0xDF,0x50,"h720"  },	/*  6 720KB AT      */
	{ 2880,18,2,80,0,0x1B,0x00,0xCF,0x6C,"H1440" },	/*  7 1.44MB 3.5"   */
	{ 5760,36,2,80,0,0x1B,0x43,0xAF,0x54,"E2880" },	/*  8 2.88MB 3.5"   */
	{ 6240,39,2,80,0,0x1B,0x43,0xAF,0x28,"E3120"},	/*  9 3.12MB 3.5"   */

	{ 2880,18,2,80,0,0x25,0x00,0xDF,0x02,"h1440" }, /* 10 1.44MB 5.25"  */
	{ 3360,21,2,80,0,0x1C,0x00,0xCF,0x0C,"H1680" }, /* 11 1.68MB 3.5"   */
	{  820,10,2,41,1,0x25,0x01,0xDF,0x2E,"h410"  },	/* 12 410KB 5.25"   */
	{ 1640,10,2,82,0,0x25,0x02,0xDF,0x2E,"H820"  },	/* 13 820KB 3.5"    */
	{ 2952,18,2,82,0,0x25,0x00,0xDF,0x02,"h1476" },	/* 14 1.48MB 5.25"  */
	{ 3444,21,2,82,0,0x25,0x00,0xDF,0x0C,"H1722" },	/* 15 1.72MB 3.5"   */
	{  840,10,2,42,1,0x25,0x01,0xDF,0x2E,"h420"  },	/* 16 420KB 5.25"   */
	{ 1660,10,2,83,0,0x25,0x02,0xDF,0x2E,"H830"  },	/* 17 830KB 3.5"    */
	{ 2988,18,2,83,0,0x25,0x00,0xDF,0x02,"h1494" },	/* 18 1.49MB 5.25"  */
	{ 3486,21,2,83,0,0x25,0x00,0xDF,0x0C,"H1743" }, /* 19 1.74 MB 3.5"  */

	{ 1760,11,2,80,0,0x1C,0x09,0xCF,0x00,"h880"  }, /* 20 880KB 5.25"   */
	{ 2080,13,2,80,0,0x1C,0x01,0xCF,0x00,"D1040" }, /* 21 1.04MB 3.5"   */
	{ 2240,14,2,80,0,0x1C,0x19,0xCF,0x00,"D1120" }, /* 22 1.12MB 3.5"   */
	{ 3200,20,2,80,0,0x1C,0x20,0xCF,0x2C,"h1600" }, /* 23 1.6MB 5.25"   */
	{ 3520,22,2,80,0,0x1C,0x08,0xCF,0x2e,"H1760" }, /* 24 1.76MB 3.5"   */
	{ 3840,24,2,80,0,0x1C,0x20,0xCF,0x00,"H1920" }, /* 25 1.92MB 3.5"   */
	{ 6400,40,2,80,0,0x25,0x5B,0xCF,0x00,"E3200" }, /* 26 3.20MB 3.5"   */
	{ 7040,44,2,80,0,0x25,0x5B,0xCF,0x00,"E3520" }, /* 27 3.52MB 3.5"   */
	{ 7680,48,2,80,0,0x25,0x63,0xCF,0x00,"E3840" }, /* 28 3.84MB 3.5"   */

	{ 3680,23,2,80,0,0x1C,0x10,0xCF,0x00,"H1840" }, /* 29 1.84MB 3.5"   */
	{ 1600,10,2,80,0,0x25,0x02,0xDF,0x2E,"D800"  },	/* 30 800KB 3.5"    */
	{ 3200,20,2,80,0,0x1C,0x00,0xCF,0x2C,"H1600" }, /* 31 1.6MB 3.5"    */
};
const int n_info = (sizeof(info_list) / sizeof(struct floppy_struct));


// Given a track, turn it into a byte count for the lseek system call
#define SEEK_LOC(cur_track) (long) (	                        \
	(long) cur_track *                                      \
	(long) (floppy_info.head * SECTOR_SIZE * floppy_info.sect)	\
    )

/********************************************************
 * raw_file -- A class to do raw file I/O		*
 *							*
 * This is a C++ class to package the open/read/write/	*
 * close system calls.					*
 *							*
 * Member functions are similar to the system calls	*
 ********************************************************/
class raw_file {
    private:
	int fd;	/* File descriptor of the raw file */
    public:
	// Create class and open file.
	raw_file(const char name[], const int flags = O_RDONLY, 
		const int prot = 0666) {
	    fd = open(name, flags, prot);
	}
	// Create class with no file
	raw_file(void) : fd(-1) {};
	// Destroy class, close file
	~raw_file(void) {
	    if (fd >= 0) {
		::close(fd);
		fd = -1;
	    }
	}
    private:
	raw_file(const raw_file &);	// No copy constructor
	raw_file & operator = (const raw_file &);// No assignment operator
    public:
	// Set if we have a problem with open or such
	bool bad(void) {
	    return (fd < 0);
	}
	// Open the file
	int open(const char name[], const int flags = O_RDONLY, 
		const int prot = 0666) {
	    if (fd >= 0) {
		::close(fd);
	        fd = -1;
	    }
	    fd = ::open(name, flags, prot);
	    return (fd);
	}
	// Read the file
	size_t read(char *data, size_t size) {
	    return (::read(fd, data, size));
	}
	size_t write(char *data, size_t size) {
	    return (::write(fd, data, size));
	}
	size_t lseek(size_t where, int how = SEEK_SET) {
	    return (::lseek(fd, where, how));
	}
	int ioctl(int op, void *param) { 
	    return (::ioctl(fd, op, param));
	}
	int fsync(void) {
	    return (::fsync(fd));
	}
	int close(void)
	{
	    // Status of the close
	    int status = ::close(fd);
	    fd = -1;
	    return (status);
	}
};

/********************************************************
 * buffer -- a data buffer class.			*
 *							*
 * This is a block of memory that is destroyed by the	*
 * destructorl.						*
 ********************************************************/
class buffer {
    private:
	char *data;
    public:
	buffer(void):data(NULL) {};
	~buffer(void) {
	    if (data != NULL) {
		delete data;
		data = NULL;
	    }
	}
    private:
	buffer(const buffer &);	// No copy constructor
	buffer &operator = (const buffer &);	// No assignment operator
    public:
	void allocate(size_t size) {
	    if (data != NULL) {
		delete data;
		data = NULL;
	    }
	    data = new char[size];
	}
	char *data_ptr(void) {
	    return (data);
	}
};

static buffer buf;	// Buffer to use

		
/********************************************************
 * prompt -- prompt the user for the next disk		*
 ********************************************************/
static void prompt(
    const char name[],		/* Name of the floppy to insert */
    const int current,		/* Current counter */
    const int total		/* Total number of floppies of this type */
)
{
    char ans[80];	/* Answer to the insert question */

    if (beeps) {
	fprintf (stderr, "");
	usleep (200000L);
	fprintf (stderr, "");
	usleep (200000L);
	fprintf (stderr, "");
    }

    printf ("%s #%d of %d: Insert floppy "
	    "into drive and press <Enter>",
	    name, current, total);
    fgets(ans, sizeof(ans), stdin);
}
/********************************************************
 * find_name -- given floppy info -- return a name	*
 ********************************************************/
const char *const find_name(
    struct floppy_struct *our_info	// The info to check
)
{
    int i;	// Random index

    for (i = 0; i < n_info; ++i) {
	if ((our_info->size == info_list[i].size) &&
	    (our_info->sect == info_list[i].sect) &&
	    (our_info->head == info_list[i].head) &&
	    (our_info->track == info_list[i].track)) {
	    return (info_list[i].name);
	}
    }
    fprintf(stderr, "Error: Could not determine floppy size\n");
    exit (8);
}
/********************************************************
 * write_protect -- See if the disk is write protected	*
 ********************************************************/
static bool write_protect(void)
{
    raw_file fd;		/* File descriptor for the device */
    /* Information about the floppy */
    struct floppy_drive_struct floppy_info;
    int probe_index;	// Index for probing.

    for (probe_index = 0; probe_list[probe_index] != NULL; ++probe_index) {
	sprintf(probe_device_name, probe_list[probe_index], device_number);

	fd.open(probe_device_name, O_RDONLY);

	if (fd.bad()) {
	    printf("Warning: %s failed errno=%d\n", probe_device_name, errno);
	    continue;
	}

	if (fd.ioctl(FDPOLLDRVSTAT, &floppy_info) != 0) {
	    printf(
		"ERROR: Can not get write protection information for floppy\n");
	    exit (8);
	}
	return ((floppy_info.flags & FD_DISK_WRITABLE) == 0);
    }
    printf("Cannot open %s for reading Error number=%d\n", 
		probe_device_name, errno);
    exit(8);
}
/********************************************************
 * set_size -- set the size/name of the device		*
 ********************************************************/
static void set_size(void)
{
    raw_file fd;		/* File descriptor for the device */

    fd.open(probe_device_name, O_RDONLY);

    if (fd.bad()) {
	printf("Cannot open %s for reading Error number=%d\n", 
			probe_device_name, errno);
	exit(8);
    }

    if (fd.ioctl(FDGETPRM, &floppy_info) != 0) {
	printf("ERROR: Can not get structure information for floppy\n");
	exit (8);
    }
    regex_t reg_info;		// Regular expression information

    regcomp(&reg_info, "^/dev/fd[01]$", 0);

    if (regexec(&reg_info, probe_device_name, 
			0, NULL, 0) == 0) {
	sprintf(actual_device_name, "%s%s", 
			probe_device_name, find_name(&floppy_info));
    } else {
	strcpy(actual_device_name, probe_device_name);
    }

    buf.allocate(floppy_info.size * 512);

    if (debug) {
	printf("Floppy %s: Size %d Sectors %d Heads %d Tracks %d Stretch %d\n",
	    probe_device_name,
	    floppy_info.size, floppy_info.sect, 
	    floppy_info.head, floppy_info.track,
	    floppy_info.stretch);
	printf("Floppy %s: Gap %d Rate %d Name: %s\n",
		probe_device_name,
		floppy_info.gap, floppy_info.rate,
		find_name(&floppy_info));
	fflush(stdout);
    }

    track_size = ((long)floppy_info.sect) * ((long)floppy_info.head) * 512L;
}
/********************************************************
 * read_disk -- read a floppy 				*
 ********************************************************/
static void read_disk(void)
{
    raw_file fd;	/* Fd of the disk we are reading */
    unsigned int cur_track;	/* Current track number */
    int read_size;	/* Size of the last read */
    char *cur_ptr;	/* Pointer to current place in the buffer */

    fd.open(actual_device_name, O_RDONLY);

    if (fd.bad()) {
	printf("Cannot open %s for reading %d\n", 
			actual_device_name, errno);
	exit(8);
    }

    cur_ptr = buf.data_ptr();

    if (graphic) {
	for (cur_track = 0; cur_track < floppy_info.track; cur_track++) {
	    putchar('.');
	}
	putchar('\r');
    }
    for (cur_track = 0; cur_track < floppy_info.track; cur_track++) {
	long seekres;	/* Location we want to seek to */

	seekres = fd.lseek(SEEK_LOC(cur_track), SEEK_SET);
	if (seekres < 0) {
	    printf ("\nSeek error %ld Error number=%d\n", seekres, errno);
	    exit(8);
	}
	if (graphic) {
	    putchar('R');
	} else {
	    printf ("\rTrack %3d Reading", cur_track);
	}
	read_size = fd.read(cur_ptr, track_size);

	if (!graphic)
	    printf ("\rTrack %3d Read   ", cur_track);

	if (read_size != track_size) {
	    printf ("\nRead error %ld != %d Error code=%d\n", 
		    track_size, read_size, errno);
	    if (!continue_on_error) {
		exit(8);
	    } else {
		/* Pointer to current location (sector) */
		char *sect_cur_ptr = cur_ptr;	

		unsigned long int cur_sector;	/* Current sector number */

		printf("\nRead track error.  Falling back to sector reads\n");

		for (cur_sector = 0; 
		     cur_sector < floppy_info.sect * floppy_info.head; 
		     ++cur_sector) {

		    seekres = fd.lseek( 
			    SEEK_LOC(cur_track) + SECTOR_SIZE  * cur_sector, 
			    SEEK_SET);
		    if (seekres < 0) {
			printf("\nSeek error %ld Error code=%d\n", 
					seekres, errno);
			exit(8);
		    }
		    printf ("\rTrack %3d Sector %ld Reading", 
			    	cur_track, cur_sector);
		    read_size = fd.read(sect_cur_ptr, SECTOR_SIZE);
		    printf ("\rTrack %3d Sector %ld Read   ", 
			    cur_track, cur_sector);
		    if (read_size != SECTOR_SIZE) {
			printf ("\nRead error %d Error number = %d\n", 
					read_size, errno);
		    }
		}
	    }
	}
	cur_ptr += track_size;
    }
    if (graphic)
	putchar('\n');

    printf ("\rFloppy read              \n");
}

/********************************************************
 * write_disk -- write out the floppy and 		*
 *	format it if needed.  Also verifies.		*
 ********************************************************/
static void write_disk(void)
{
    raw_file fd;	/* File descriptor of the drive */
    int cur_track;	/* Current track number */
    char *cur_ptr;	/* Pointer to what we're writing now */
    long seekres;	/* Location we want to seek to */
    buffer verify_buffer; /* Buffer for verification */
    int write_size;	/* Size of the last write */
    int status;		/* Status of the last ioctl */
    int read_size;	/* Size of the last read */

    verify_buffer.allocate(track_size);
    fd.open(actual_device_name, O_RDWR);
    if (fd.bad()) {
	printf ("Cannot open %s for writing. Error number=%d\n", 
		    actual_device_name, errno);
	exit(8);
    }
    static struct format_descr fmtbuf;

    cur_ptr = buf.data_ptr();
    if (graphic) {
	for (cur_track = 0; cur_track < int(floppy_info.track); cur_track++) {
	    putchar('.');
	}
	putchar('\r');
    }

    for (cur_track = 0; cur_track < int(floppy_info.track); cur_track++) {

	if (format) {
	    unsigned int cur_head;	/* Current head to format */

	    for (cur_head = 0; 
		 cur_head < floppy_info.head;
		 ++cur_head) {
		int status;	/* Status of the ioctl */

		if (graphic) {
		    putchar( (cur_head == 0) ? 'f' : 'F' );
		    putchar('\b');
		} else {
		    printf ("\rTrack %3d Format head %d", cur_track, cur_head);
		}
		fmtbuf.track = cur_track;
		fmtbuf.head = cur_head;
		status = fd.ioctl(FDFMTTRK, &fmtbuf);
		if (status < 0) {
		    printf(
			"\nIoctl FDFMTTRK error status = %d error code = %d\n", 
			status, errno);
		    return;
		}
	    }
	} /* End of track formatting */

	seekres = fd.lseek(SEEK_LOC(cur_track), SEEK_SET);
	if (seekres < 0) {
	    printf ("\nSeek error %ld Error number = %d\n", seekres, errno);
	    return;
	}
		    	
	if (graphic) {
	    putchar('w');
	} else {
	    printf ("\rTrack %3d Writing      ", cur_track);
	}

	write_size = fd.write(cur_ptr, track_size);
	if (write_size != track_size) {
	    printf ("\nWrite error %d error code=%d\n", write_size, errno);
	    return;
	}
	fd.fsync();

	if (!graphic)
	    printf ("\rTrack %3d Written      ", cur_track);

	cur_ptr += track_size;
    }
    status = fd.ioctl(FDFLUSH, NULL);
    if (status < 0) {
	printf ("\nIoctl FDFLUSH error %d %d\n", status, errno);
	return;
    }
    if (verify) {
	sleep (2);
	for (cur_track = floppy_info.track - 1; cur_track >= 0; cur_track--) {
	    cur_ptr -= track_size;

	    seekres = fd.lseek(SEEK_LOC(cur_track), 0);
	    if (seekres < 0) {
		printf ("\nSeek error %ld error number=%d\n", seekres, errno);
		return;
	    }
	    if (graphic) {
		putchar('\b');
		putchar('v');
		putchar('\b');
	    } else {
		printf ("\rTrack %3d Reading ", cur_track);
	    }

	    read_size = fd.read(verify_buffer.data_ptr(), track_size);
	    if (!graphic) {
		printf ("\rTrack %3d Read    ", cur_track);
	    }

	    if (read_size != track_size) {
		printf ("\nRead error %d error number=%d\n", read_size, errno);
		return;
	    }

	    if (memcmp(cur_ptr, verify_buffer.data_ptr(), track_size) == 0) {
		if (graphic) {
		    putchar('V');
		    putchar('\b');
		} else { 
		    printf ("\rTrack %3d Verified", cur_track);
		}
	    } else {
		printf ("\nMiscompare at track %d\n", cur_track);
		return;
	    }
	}
    }
    printf ("\nFloppy ");
    if (format) {
	printf ("formatted and ");
    }
    printf ("written ");
    if (verify) {
	printf ("and verified");
    }
    printf ("                   \n");
}

int main (int argc, char *argv[])
{
    int cur_source;
    int cur_dest;		/* Current destination disk */
    int i;			/* Current getopt character */

    while ((i = getopt (argc, argv, "Dcbvfd:s:Vg")) != EOF) {
	switch (i) {
	    case 'D':
		device_number = 1;	// Number of the device
		break;
	    case 'g':		/* Graphic progress display */
		graphic = true;
		break;	
	    case 'V':
		debug = true;
		break;
	    case 'b':		/* Beep when changing disks */
		beeps = true;
		break;
	    case 'f':		/* Format destination */
		format = true;
		break;
	    case 'v':		/* Verify disk */
		verify = true;
		break;
	    case 'd':		/* # of disks to write */
		dst_count = atoi (optarg);
		break;
	    case 's':
		src_count = atoi (optarg);
		break;
	    case 'c':		/* # Continue on error */
		continue_on_error = 1;
		break;
	    default:
		printf ("Usage: fdup [-fvcbDS] [-s n] [-d n]\n");
		printf ("   -f    format destination\n");
		printf ("   -v    verify write\n");
		printf ("   -c    continue on read error\n");
		printf ("   -b    beep at diskette change\n");
		printf ("   -D    use drive 1 (default: drive 0)\n");
		printf ("   -S    drive is 1.44 meg (default: 1.2 meg)\n");
		printf ("   -s n  number of source diskettes (default: 1000)\n");
		printf ("   -d n  number of destination diskette sets (default: 1)\n");
		printf ("   -V    Turn on verbose messages\n");
		printf ("   -g    Display progress graphically\n");
		exit(8);
		break;
	}
    }
    setvbuf (stdout, NULL, _IONBF, 0);
    setvbuf (stderr, NULL, _IONBF, 0);
    printf ("Floppy duplicator.  V3.0.  by Steve Oualline\n");
    printf("(from an original work by Mark K Vallevand)\n");
    printf ("Destination floppies will be ");
    if (format) {
	printf ("formatted & ");
    }
    printf ("written ");
    if (verify) {
	printf ("& verified ");
    }
    printf ("in drive\n");

    for (cur_source = 1; cur_source <= src_count; cur_source++) {
	while (1) {
	    prompt("Source", cur_source, src_count);
	    if (write_protect())
		break;
	    printf("Error: Source disk is not write protected\n");
	}
	set_size();
	read_disk();

	for (cur_dest = 1; cur_dest <= dst_count; cur_dest++) {
	    while (1) {
		prompt("Destination", cur_dest, dst_count);
		if (!write_protect())
		    break;
		printf("Error: Destination disk is write protected\n");
	    }
	    write_disk();
	}
    }
    return (0);
}
